############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import time, warnings, types, weakref

import ZSI.TC
from ZSI import  _seqtypes
from ZSI.TC import _get_global_element_declaration as GED
from ZSI.TC import ElementDeclaration
from ZSI.writer import SoapWriter
from ZSI.parse import ParsedSoap
from ZSI.wstools.Namespaces import OASIS
from ZSI.wstools.Utility import SplitQName

from twisted.python import log
from twisted.internet import reactor
from ZSI.wstools.Utility import SplitQName

#from pyGridWare.faults import *
from pyGridWare.faults.PropertiesFaults import *


# Set to one year.
DEFAULT_TERMINATION = 365*24*60*60

'''
Classes:
    NoResourcePropertyError
    DeclarationError
    ResourceProperties
    NotificationProducerRP
    ScheduledResourceTerminationRP
    SubscriptionManagerRP
'''


class NoResourcePropertyError(AttributeError):
    '''No Such Resource Property exists in ResourceProperty declararion.
    '''
class DeclarationError(NotImplementedError):
    '''delcaration not declared.
    '''

class ResourceProperties(object):
    '''ResourceProperties must be a GED, and it contains any number 
    references to other GEDs. This class must be subclassed, and
    declaration must be specified.

    class variables
        declaration -- must be a class type representing a GED.
    '''
    declaration = None

    def __init__(self):
        self.CheckTypecode()
        self.topicListeners = []
        self._typecode = self.declaration()
        self._pyobj = self._typecode.pyclass()

    def __str__(self):
        sw = SoapWriter(header=False, envelope=False)
        sw.serialize(self._pyobj, self._typecode)
        return str(sw)

    def setUp(self, xml=None):
        '''xml -- string representation of RP XML instance.
        '''
        if xml is None: return
        ps = ParsedSoap(xml, envelope=False)
        self._pyobj = ps.Parse(self._typecode)

    def getContext(self, context):
        return self._context()
    def setContext(self, context):
        self._context = weakref.ref(context)
    context = property(getContext, setContext, None, 'Set properties resource context')

    def getResourcePropertiesDocument(self):
        '''ResourceProperties must be a global element declaration.
        '''
        return self._pyobj

    def getResourceProperty(self, namespaceURI, localName):
        '''Resource must be a global element declaration, 
        and so assume that it is either the ResourceProperty
        element itself, or one of the element references in
        its model group.
        '''
        tc = self._typecode
        if tc.nspname == namespaceURI and  tc.pname == localName:
            log.debug("TOP-WHAT(%s, %s): " %(tc.nspname,tc.pname))
            return tc

        if hasattr(tc, 'ofwhat'):
            for what in tc.ofwhat:
                log.debug("WHAT(%s, %s): " %(what.nspname,what.pname))
                if what.nspname == namespaceURI and  what.pname == localName:
                    element = getattr(self._pyobj, what.aname, None)
                    log.debug("ELEMENT: ", element)

                    if element is not None:
                        if type(element) is types.InstanceType:
                            element.typecode = what
                            return element
                        else:
                            class Wrapper(type(element)):
                                typecode = what
                            return Wrapper(element)
                    return None

        raise InvalidResourcePropertyQNameFault(timestamp=time.gmtime(),
            faultcause=('no element (%s,%s) in resourceProperty(%s,%s) '\
            %(namespaceURI,localName,tc.nspname,tc.pname),),
        )

    def setResourceProperty(self, namespaceURI, localName, pyobj):
        what = self._typecode
        for what in what.ofwhat:
            if what.nspname == namespaceURI and  what.pname == localName:
                break
        else:
            what = self._typecode
            raise SetResourcePropertyRequestFailedFault(timestamp=time.gmtime(),
                faultcause=('no element (%s,%s) in resourceProperty(%s,%s) '\
                %(namespaceURI,localName,what.nspname,what.pname),),
            )

        setattr(self._pyobj, what.aname, pyobj)
        log.debug('--> TopicListeners: %s' %self.topicListeners)
        if pyobj is None: 
            if what.minOccurs == 0 or what.nillable:
                if len(self.topicListeners) > 0:
                    warnings.warn(\
                        'Not notifying TopicListeners of change of property(%s,%s) to None'\
                        %(what.nspname,what.pname)
                    )
                return
            warnings.warn('missing instance of class %s minOccurs=%d, nillable=%d' \
                %(what.__class__, what.minOccurs, what.nillable)
            )
            raise SetResourcePropertyRequestFailedFault(timestamp=time.gmtime(),
                faultcause=('missing instance of element(%s, %s)' %(what.nspname, what.pname)),
            )
        elif what.maxOccurs == 1:
            if hasattr(pyobj, 'typecode') is False:
                try: 
                    pyobj.typecode = what
                except:
                    pyType = type(pyobj)
                    if pyType is bool: pyType = int

                    class Wrapper(pyType):
                        typecode = what
                    pyobj = Wrapper(pyobj)

        elif type(pyobj) in _seqtypes:
            for indx in range(len(pyobj)):
                obj = pyobj[indx]
                if hasattr(obj, 'typecode') is False:
                    try: 
                        obj.typecode = what
                    except:
                        pyType = type(obj)
                        if pyType is bool: pyType = int

                        class Wrapper(pyType):
                            typecode = what
                        obj = Wrapper(obj)

                        pyobj[indx] = obj
                else:
                    break
        else:
            raise TypeError, 'element (%s,%s) is repeatable must be a sequence' \
                %(what.nspname,what.pname)

        #
        # Need to time out on callbacks.
        #
        for cb in self.topicListeners:
            log.debug('--> Callback')
            reactor.callLater(0.0001, cb, pyobj, namespaceURI, localName)

    def CheckTypecode(cls, typecode=None):
        '''Recursive base check to make sure declaration is a superset of all declarations 
        of RP subclasses.  If typecode is None, grab the class variable typecode and
        check it against its base classes. 
        '''
        ctypecode = cls.GetTypecode()
        if ctypecode is None: 
            return

        for base in filter(lambda b: issubclass(b, ResourceProperties), cls.__bases__):
            base.CheckTypecode(typecode or ctypecode)

        if typecode is None: 
            return

        ofwhat = map(typecode.ofwhat, lambda w: (w.nspname,w.pname))
        for what in ctypecode.ofwhat:
            if (what.nspname,what.pname) not in ofwhat:
                raise ResourceUnknownFault(timestamp=time.gmtime(),
                    faultcause=('RP declaration must declare (%s,%s)' %(what.nspname,what.pname),),
                )
    CheckTypecode = classmethod(CheckTypecode)

    def GetTypecode(cls):
        if getattr(cls, 'declaration', None) is None:
             return None
        if issubclass(cls.declaration, ElementDeclaration) is False:
             raise DeclarationError, 'declaration class variable must be an ElementDeclaration.'
        return cls.declaration()
    GetTypecode = classmethod(GetTypecode)


class NotificationProducerRP(ResourceProperties):
    '''NotificationProducer Resource Properties.
    Topic 
    FixedTopicSet 
    TopicExpressionDialects 
    '''
    declaration = GED(OASIS.BASENOTIFICATION, 'NotificationProducerRP')

    def __init__(self):
        pass

    def setUp(self, xml=None):
        if xml:
            ps = ParsedSoap(xml, envelope=False)
            self._pyobj = ps.Parse(self._typecode)
            return

        self.FixedTopicSet = True
        self.TopicExpressionDialects = ["http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple",]

        tc = None
        for what in self._pyobj.typecode.ofwhat:
            if (what.nspname,what.pname) == (OASIS.BASENOTIFICATION,"Topic"):
                   tc = what
                   break

        assert tc is not None, \
            'Declaration for ResourceProperty: Expecting element Topic for NotificationProducer'

        l = []
        class QName(tuple):
            typecode = ZSI.TC.QName()

        for what in self._pyobj.typecode.ofwhat:
            pyobj = tc.pyclass()
            setattr(pyobj, tc.mixed_aname, QName((what.nspname,what.pname)))
            l.append(pyobj)

        self.Topic = l

    def getTopic(self):
        return self.getResourceProperty(OASIS.BASENOTIFICATION, "Topic")
    def setTopic(self, value):
        self.setResourceProperty(OASIS.BASENOTIFICATION, "Topic", value)
    Topic = property(getTopic, setTopic, None, 'Topic')

    def getFixedTopicSet(self):
        return self.getResourceProperty(OASIS.BASENOTIFICATION, "FixedTopicSet")
    def setFixedTopicSet(self, value):
        self.setResourceProperty(OASIS.BASENOTIFICATION, "FixedTopicSet", value)
    FixedTopicSet = property(getFixedTopicSet, setFixedTopicSet, None, 'Fixed Topic Set')

    def getTopicExpressionDialects(self):
        return self.getResourceProperty(OASIS.BASENOTIFICATION, "TopicExpressionDialects")
    def setTopicExpressionDialects(self, value):
        self.setResourceProperty(OASIS.BASENOTIFICATION, "TopicExpressionDialects", value)
    TopicExpressionDialects = property(getTopicExpressionDialects, setTopicExpressionDialects, None, \
        'Topic Expression Dialects')

    def addTopicListener(self, method, topic):
        ''' Need to hold onto DOM to make this really work
        specification is retarded, uses text content to pass
        in a qualifiedName.
        '''
        assert callable(method), 'Callback must be callable.'
        if not getattr(topic, '_text', None): return
        prefix,name = SplitQName(topic._text)
        log.debug("TOPIC: %s:%s" %(prefix,name))

        def cb(value, nspname, pname):
            if pname != name: return
            return method(value, topic)

        #self.topic = topic
        self.topicListeners.append(cb)



class ScheduledResourceTerminationRP(ResourceProperties):
    '''
                    self._CurrentTime = None
                    self._TerminationTime = None
    '''
    declaration = GED(OASIS.BASENOTIFICATION, 'ScheduledResourceTerminationRP')

    def __init__(self):
        self.currentTime = time.gmtime()
        self.terminationTime = time.gmtime(time.time()+DEFAULT_TERMINATION)

    def getCurrentTime(self):
        return self.getResourceProperty(OASIS.LIFETIME, "CurrentTime")
    def setCurrentTime(self, value):
        assert isinstance(value, tuple) or isinstance(value, time.struct_time) and len(value) == 9,\
            'expecting 9 item tuple got len(%d), %s' %(len(value),str(value))
        self.setResourceProperty(OASIS.LIFETIME, "CurrentTime", tuple(value))
    currentTime = property(getCurrentTime, setCurrentTime, None, 'current time')

    def getTerminationTime(self):
        return self.getResourceProperty(OASIS.LIFETIME, "TerminationTime")
    def setTerminationTime(self, value):
        assert isinstance(value, tuple) or isinstance(value, time.struct_time) and len(value) == 9,\
            'expecting 9 item tuple got len(%d), %s' %(len(value),str(value))
        self.setResourceProperty(OASIS.LIFETIME, "TerminationTime", tuple(value))
    terminationTime = property(getTerminationTime, setTerminationTime, None, 'Termination time')



class SubscriptionManagerRP(ScheduledResourceTerminationRP):
    '''
    CurrentTime 
    TerminationTime 
    ConsumerReference 
    TopicExpression 
    UseNotify 
    Precondition 
    Selector 
    SubscriptionPolicy 
    CreationTime 
    '''
    declaration = GED(OASIS.BASENOTIFICATION, 'SubscriptionManagerRP')

