#!/bin/env python
############################################################################
# Keith Beattie <KSBeattie@lbl.gov>, LBNL
# See Copyright for copyright notice!
# $Id: gridSniffer.py,v 1.12 2003/12/12 03:02:43 boverhof Exp $
###########################################################################
import exceptions, getopt, re, sys, socket, time, threading, urlparse
try:
     from xml.ns import WSDL
except Exception,ex:
    BINDINGS = {}
else:
    # Binding Dictionary
    BINDINGS = {
        'soap':WSDL.BIND_SOAP,
        'http':WSDL.BIND_HTTP,
        'mime':WSDL.BIND_MIME,
        'all':WSDL.BASE
}
BINDING = None
ALLHOSTNAMES = True
REPLACE_ALL = False

"""

GridSniffer:

A tcpmon-style 'sniffer' logging the upstream and downstream messages
exchanged between a tcp client and server.

URLs of the form 'http://<serverIP>:<serverPort>/' (where <serverIP>
is the IP of the server) are re-written to
'http://<localIP>:<listenPort>/', In this way we stay in the loop on
messages exachanged.  This is the default which can be disabled with
the '-r' flag.


Use:

The expected use of this is to be run in it's own terminal, where it
will print out information about connections made, the amount of data
being exchanged the number of rewrites done, etc.

To terminate it, interrupt it with a Cntl-C once it reaches the
'looping to accept new client' state. (Otherwise there is still a
thread running and it will take a kill to terminate the python
interpreter).


Arch:

This works by creating two threads, one for listening to upstream
(client to server) data and another for listening to downstream data
(server to client).

GridSniffer loops forever accepting connections from client.  Upon accept:
  - Connects to server
  - Spawns upstream and downstream threads
  - Waits for them to finish

UpStream thread:
  - reads data from client
  - prints it to log
  - sends it to server
  - when connection is closed, shutdown downstream (client) socket.

DownStream thread:
  - reads data from server
  - possibly re-writes URLs
  - prints it to log
  - sends it to client
  - when connection is close, shutdown upstream (server) socket.


Wish List (in no particular order):

  - Now that there are _getHTTPHeader and _getHTTPBody methods it
  could be possible to have a single log file, rather than two.  This
  would require making the upstream thread more HTTP aware and then
  collating messages between the threads into request-response pairs
  and writing them out into a single log file (probably another
  thread, or perhaps this would all be more easily done using the
  select-model...).  My guess. though. is that my assumptions about
  the exact structure of HTTP messages is not perfect.

  - Expand the regexp for re-writes to replace more than just the
  'http://serverIP:serverPort/' case (though that's the only one I've
  seen from GRID containers).  Though HTTP isn't a priority here, it
  doesn't handle '301 Moved Permanently' warnings.

  - If the kinds of re-writes that are done is expanded (as suggested
  above) consider breaking the re-writing out to a
  DownStream.rewrite() method as the repl arg of the re.subn() call.
  This way details of what is being re-written can be printed to
  stdout.

  - Is it possible that a redirect is not an indication that the
  current connection will be closing very soon?  I assume that if
  there is a re-direct, then the current connection will be shutdown
  and a new one will be initiated by the client.  Perhaps it would be
  better to spawn a whole new GridSniffer thread for each re-write or
  each connection?  Ugh, that'd be a big change and it wold mess up
  the logging.  This is beyond what I think this tool is going to be
  used for.

  - Speaking of big changes, it might be better to switch to the
  'select model' (as opposed to the current 'thread model') and allow
  multiple accepts to be open at once.  Then we'd wait on messages from any
  of the sockets and track (exactly how I'm not sure yet) which one
  sends to which.  This would be a better strategy if we need to
  maintain multiple connections to servers in order to stay
  'in-the-middle'.


"""

# Usage string
USAGE = """Usage: gridSniffer -l listenPort -s server [-p serverPort] [-u upLogFile] [-d downLogFile] [-a] [-b soap,mime,http,all] [-r] [-x] [-h]
  where:
    listenPort  -> The port gridSniffer is listening on (the one your client connects to).
    server      -> The server running the grid container.
    serverPort  -> The port the server is listening on [80].
    upLogFile   -> The file to write the upstream log to [./gs_up.log].
    downLogFile -> The file to write the downstream log to [./gs_down.log].
    -a          -> Append to log files rather than overwirte them.
    -b          -> Check namespace of bindings, %s
    -r          -> Don't re-write URLs from server:port to localhost:listenPort.
    -x          -> Re-write all occurances of http://hostname:port to http://hostname:listenport
    -h          -> prints this message and exits.
$Id: gridSniffer.py,v 1.12 2003/12/12 03:02:43 boverhof Exp $""" %BINDINGS.keys()

class GridSnifferException(exceptions.Exception):
    """ A GridSniffer specific exception class """

    def __init__(self, args=None):
        """ Generic exception construction """
        exceptions.Exception.__init__(self, args)


class UpStream(threading.Thread):

    """
    The up-stream thread reading from the client printing to the log
    and sending to the server.
    """

    def __init__(self, csock, ssock, log):

        """
        Construct an UpStream object with two opened sockets and an
        open log file.

        @type  csock: socket
        @param csock: An open socket to the client

        @type  ssock: socket
        @param ssock: An open socket to the server

        @type  log: file
        @param log: An open log file

        """

        threading.Thread.__init__(self)
        self.__clientSock = csock
        self.__serverSock = ssock
        self.__logFile    = log

        # Log timestamp, etc. about new connection
        now = time.time()
        peer = self.__clientSock.getpeername()
        print >>self.__logFile, \
          "\n===== GridSniffer: Connection from %s:%s at %s:%08.5f =====" % (
            peer[0], peer[1], time.strftime('%Y %a %b %d %H:%M',
                                            time.localtime(now)), now % 60)

    def run(self):

        """
        Read from client, log it and send it to server.  Return when
        the client socket is closed, but first shutdown the server
        socket as a notification to the DownStream thread to finish.
        """

        try:
            while 1:
                data = self.__clientSock.recv(8192)
                print "gs.UpStream: read %s bytes" % (len(data))
                if not data:
                    self.__serverSock.shutdown(2)  # This makes the DownStream thread stop
                    break
                print >>self.__logFile, data,
                self.__logFile.flush()
                self.__serverSock.sendall(data)
        except socket.error, e:
            print "gs.UpStream.run(): socket.error:", e
            print sys.exc_info()[2]    # Traceback


class DownStream(threading.Thread):

    """
    The down-stream thread reading from the server printing, doing
    re-writes, printing to the log and sending to the client.
    """

    def __init__(self, csock, ssock, log, rewrite, local, server):

        """
        Construct a DownStream object from two open sockets, an
        open log file, the rewrite regular expression, our local IP
        address and the port that we are listening on.

        @type  csock: socket
        @param csock: An open socket to the client

        @type  ssock: socket
        @param ssock: An open socket to the server

        @type  log: file
        @param log: An open log file

        @type  rewrite: tuple(string, string)
        @param rewrite: A tuple of: The search string to
                        look for, the replacement string

        @type  local: tuple(sting, number)
        @param local: A tuple of: Our local IP address and the port we are
                      listening on (used for rewriting URLs)

        @type  server: tuple(sting, number)
        @param server: A tuple of: Our server IP address and the port we are
                      sending to
        """

        threading.Thread.__init__(self)
        self.__clientSock = csock
        self.__serverSock = ssock
        self.__logFile    = log
        self.__rewriteRE  = []
        self.__searchStr, self.__replaceStr = rewrite
        self.__localIP, self.__listenPort = local
        self.__serverIP,self.__serverPort  = server

        self.__replaceStr = self.__replaceStr %\
                (self.__localIP,self.__listenPort)

        # Log timestamp, etc. about new connection
        now = time.time()
        peer = self.__serverSock.getpeername()
        print >>self.__logFile, \
          "===== GridSniffer: Connection to %s:%s at %s:%08.5f =====" % (
          peer[0], peer[1], time.strftime('%Y %a %b %d %H:%M',
                                          time.localtime(now)), now % 60)

    def run(self):
        """
        Read from server, log it and send it to client.  Return when
        the server socket is closed, but first shutdown the client
        socket as a notification to the UpStream thread to finish.
        """
        searchStrings = []
        cname,aliass,ips = socket.gethostbyaddr(self.__serverIP)
        check_hostnames = [cname]

        if ALLHOSTNAMES:
            check_hostnames += ips
            check_hostnames += aliass
        else:
            check_hostnames += self.__serverIP

        for addr in  check_hostnames:
            searchStrings.append(\
               self.__searchStr %(addr, self.__serverPort))

        if not BINDING and not len(self.__rewriteRE):
            for search in searchStrings:
                self.__rewriteRE.append(re.compile(search))

        try:
            serverSockFile = self.__serverSock.makefile()
            while 1:  # Server socket it open, one loop per response

                # Get headers and message body
                header, headerLength, contentLength = self._getHTTPHeader(serverSockFile)
                if contentLength == 0:  # socket read 0 => socket closed
                    self.__clientSock.shutdown(2)  # To stop the UpStream thread
                    break
                body = self._getHTTPBody(serverSockFile, contentLength)
                print "gs.DownStream: read %s bytes" % (headerLength + contentLength)

                # Do re-writes, log and send to client
                headerStr = ''.join(header)
                bodyStr = ''.join(body)

                if self.__searchStr:
                    count = 0
                    if REPLACE_ALL:
                        #netlocR = '%s:%d' %(self.__localIP, self.__serverPort)
                        netlocR = '%s:%d' %('localhost', self.__listenPort)
                        for addr in check_hostnames:
                            netlocS = '%s:%d' %(addr, self.__serverPort)
                            urlS = urlparse.urlunparse(('http',netlocS,'/',None,None,None))
                            urlR = urlparse.urlunparse(('http',netlocR,'/',None,None,None))
                            print "SEARCH:  ", urlS
                            print "REPLACE: ", urlR
                            bodyStr.replace(urlS, urlR)
                    elif not BINDING:
                        for rexp in self.__rewriteRE:
                            (bodyStr, subcount) = rexp.subn(self.__replaceStr,\
                                                          bodyStr)
                            count += subcount
                    else: 
                        if BINDING == 'all':
                            prefix_re = re.compile(r'xmlns:\S+\s*=\s*"%s\S+"' 
                                                    %BINDINGS[BINDING])
                            xmlns_re = re.compile(r'xmlns\s*=\s*"%s\S+"'\
                                                    %BINDINGS[BINDING])
                        else:
                            prefix_re = re.compile(r'xmlns:\S+\s*=\s*"%s"'\
                                                    %BINDINGS[BINDING])
                            xmlns_re = re.compile(r'xmlns\s*=\s*"%s"'\
                                                    %BINDINGS[BINDING])

                        for line in prefix_re.findall(bodyStr):
                            prefix = line.split('=')[0].split(':')[1]
                            replaceStr = '%s:%s' %(prefix, self.__replaceStr)
                            for searchStr in searchStrings:
                                regexpSearch = '%s:%s' %(prefix, searchStr)
                                c = re.compile(regexpSearch)
                                (bodyStr, subcount) = c.subn(replaceStr, bodyStr)
                                count += subcount

                        if len(xmlns_re.findall(bodyStr)) >= 1:
                            replaceStr = '<' + self.__replaceStr
                            for searchStr in searchStrings:
                                regexpSearch = '<' + searchStr
                                c = re.compile(regexpSearch)
                                (bodyStr, subcount) = c.subn(replaceStr, bodyStr)
                                count += subcount

                    print "gs.DownStream: %d re-writes" % (count)

                print >>self.__logFile, headerStr, bodyStr,
                self.__logFile.flush()
                self.__clientSock.sendall(headerStr)
                self.__clientSock.sendall(bodyStr)
        except socket.error, e:
            print "gs.DownStream: socket.error:", e
            print sys.exc_info()[2]    # Traceback

        serverSockFile.close()

    def _getHTTPHeader(self, serverSockFile):

        """
        Read the header of an HTTP response returning: it as a list of
        stings, total size of header and the content length of the
        body.

        @type  serverSockFile: socket.makefile()
        @param serverSockFile: The socket File to read from

        @rtype:  list (list of strings, number, number)
        @return: (The headers of the response as a list of stings,
                  the total size of the header, the Content-Length)
        """

        header = []
        contentLength = 0
        ret = (None, 0, 0)
        while 1:
            line = serverSockFile.readline()
            if not line:  # Socket closed
                if len(header):  # We're read something already...
                    raise GridSnifferException, "Socket closed while reading headers."
                break

            header.append(line)

            # Look for Content-Length header
            if line.startswith('Content-Length: '):
                contentLength = int(line.split(': ', 1)[1])
            if line == '\n' or line == '\r\n': # newline alone means end of headers
                ret = (header, len(''.join(header)), contentLength)
                break
        return ret


    def _getHTTPBody(self, serverSockFile, contentLength):

        """
        Read the body of an HTTP response into a list of stings.

        @type  serverSockFile: socket.makefile()
        @param serverSockFile: The socket File to read from

        @type  contentLength: number
        @param contentLength: The size of the body to read

        @rtype:  list of strings
        @return: The body of the response as a list of stings
        """

        body = []
        bodylen = 0

        if contentLength == 0:
            print "gs.DownStream._getBody(): 0 Content-Length!?"
            return body

        while bodylen < contentLength:
            line = serverSockFile.readline()
            if not line:
                raise GridSnifferException, "Socket closed while reading body."
            body.append(line)
            bodylen += len(line)

        return body


class GridSniffer:

    """
    Management class for the upstream and downstream threads.  Opens
    and closes, log files and sockets.
    """

    def __init__(self, listenPort, server, logFileNames, logFileMode, rewrite):

        """
        Construct a GridSniffer management object.

        @type  listenPort: number
        @param listenPort: The port we are listening on.

        @type  server: tuple(sting, number)
        @param server: A tuple of: the server to connect to and
                       the port on the server to connect to

        @type  logFileNames: tuple(sting, string)
        @param server: A tuple of: the upstream log filename and
                       the downstream log filename.

        @type  logFileMode: string
        @param logFileMode: The mode flags for opening the log files
                            either overwrite 'w' or append 'a'.

        @type  rewrite: boolean
        @param rewrite: 1 if rewrite are enabled, 0 if not.

        """

        self.__listenPort        = listenPort
        self.__serverName        = server[0]
        self.__serverPort        = server[1]
        self.__upStreamLogFile   = logFileNames[0]
        self.__downStreamLogFile = logFileNames[1]
        self.__logFileMode       = logFileMode
        self.__regexpReplace     = None
        self.__regexpSearch      = None

        # Get our and the server's IP
        self.__localIP  = socket.gethostbyname(socket.gethostname())
        self.__serverIP = socket.gethostbyname(self.__serverName)

        self.__gssock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__gssock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__gssock.bind(('', listenPort))
        self.__gssock.listen(5)

        # Calc the regular expression for URL re-writing
        if rewrite:
            self.__regexpReplace = ' location="http://%s:%d/'
            self.__regexpSearch = ' location="http://%s:%d/' 

    def go(self):

        """
        Main GridSniffer event loop.

        Loop accepting connections from clients.  Upon accect
          - connect to server
          - spawn threads
          - wait for them to finish
        """

        # open log files
        uplog = downlog = open(self.__downStreamLogFile, self.__logFileMode)
        if self.__upStreamLogFile != self.__downStreamLogFile:
            uplog   = open(self.__upStreamLogFile,   self.__logFileMode)

        try:
            while 1:
                # Wait for incomming connection:
                csock,caddr = self.__gssock.accept()
                now = time.time()
                print "gs: Connection from %s:%s %s:%08.5f" % \
                      (caddr[0], caddr[1], time.strftime('%Y %a %b %d %H:%M',
                                                         time.localtime(now)),
                      now % 60)

                # Connect to server
                ssock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                ssock.connect((self.__serverName, self.__serverPort))

                # Start upstream thread
                upstream = UpStream(csock, ssock, uplog)
                upstream.start()

                # Start downstream thread
                downstream = DownStream(csock, ssock, downlog, 
                                (self.__regexpSearch,self.__regexpReplace),
                                (self.__localIP,self.__listenPort), 
                                (self.__serverIP,self.__serverPort))
                downstream.start()

                # Wait for threads to finish (and they both will
                # because if one does it shuts down the other's socket
                # and eventually one of their sockets will be shutdown
                # via the other end.)
                downstream.join()
                upstream.join()

                # Close sockets
                csock.close()
                ssock.close()

                print "gs: Connection closed, looping to accept new client."


        except KeyboardInterrupt:
            pass

        # Close log files
        uplog.close()
        downlog.close()

        # Close listen socket
        self.__gssock.close()


#### End of class definitions and begining of main program ####


def main():

    """
    GridSniffer entry point.  Collect command line args and start the
    GridSniffer.
    """
    
    # Check command line args
    try:
        opts, args = getopt.getopt(sys.argv[1:], "l:s:p:u:d:ab:rxh")
    except getopt.GetoptError:
        print usage
        sys.exit(1)

    listenPort = serverName = None
    serverPort = 80
    upLogFile = 'gs_up.log'
    downLogFile = 'gs_down.log'
    rewrite     = 1
    logFileMode = 'w'
    try:
        for o, a in opts:
            if o == '-l':
                listenPort = int(a)
            elif o == '-s':
                serverName = a
            elif o == '-p':
                serverPort = int(a)
            elif o == '-u':
                upLogFile = a
            elif o == '-d':
                downLogFile = a
            elif o == '-a':
                logFileMode = 'a'
            elif o == '-b':
                global BINDING
                BINDING = a
            elif o == '-x':
                global REPLACE_ALL
                REPLACE_ALL = True
            elif o == '-r':
                rewrite = 0
            elif o == '-h':
                print usage
                sys.exit()
    except ValueError:
        print usage
        sys.exit(1)

    if listenPort == None or serverName == None:
        print usage
        sys.exit(1)

    # Start the GridSniffer
    gs = GridSniffer(listenPort, (serverName, serverPort),
                     (upLogFile, downLogFile), logFileMode, rewrite)
    print "GridSniffer: %d <-> %s:%d" % (listenPort, serverName, serverPort)
    gs.go()

if __name__ == "__main__":  main()
