############################################################################
# Monte M. Goode, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################

# contains text container classes for new generation generator

# $Id: containers.py,v 1.1.2.24 2005/04/20 22:48:26 boverhof Exp $
import types
from utility import StringWriter, TextProtect, TextProtectAttributeName
from utility import NamespaceAliasDict as NAD, NCName_to_ClassName as NC_to_CN

import ZSI
from ZSI.TC import _is_xsd_or_soap_ns
from ZSI.wstools import XMLSchema, WSDLTools
from ZSI.wstools.Namespaces import SCHEMA, SOAP, WSDL
from ZSI.wstools.Utility import Base
from ZSI.typeinterpreter import BaseTypeInterpreter
from ZSI.generate import WSISpec, WSInteropError, Wsdl2PythonError, WsdlGeneratorError

ID1 = '    '
ID2 = 2*ID1
ID3 = 3*ID1
ID4 = 4*ID1
ID5 = 5*ID1
ID6 = 6*ID1

DEC = '_Dec'
DEF = '_Def'

def GetClassNameFromSchemaItem(item):
    '''
    '''
    assert isinstance(item, XMLSchema.XMLSchemaComponent), 'must be a schema item.'
    alias = NAD.getAlias(item.getTargetNamespace())
    if item.isDefinition() is True:
        return '%s.%s' %(alias, NC_to_CN('%s_Def' %item.getAttributeName()))
    return None

def FromMessageGetSimpleElementDeclaration(message):
    '''If message consists of one part with an element attribute,
    and this element is a simpleType return a string representing 
    the python type, else return None.

    '''
    assert isinstance(message, WSDLTools.Message), 'expecting WSDLTools.Message'

    if len(message.parts) == 1 and message.parts[0].element is not None:
       part = message.parts[0]
       nsuri,name = part.element
       wsdl = message.getWSDL()
       types = wsdl.types
       if types.has_key(nsuri) and types[nsuri].elements.has_key(name):
            e = types[nsuri].elements[name]
            if isinstance(e, XMLSchema.ElementDeclaration) is True and e.getAttribute('type'):
                typ = e.getAttribute('type')
                bt = BaseTypeInterpreter()
                ptype = bt.get_pythontype(typ[1], typ[0])
                return ptype

    return None


class ContainerError(Exception):
    pass

class ContainerBase(Base):
    '''Base class for all Containers.
    '''
    def __init__(self):
        Base.__init__(self)
        self.content = StringWriter('\n')
        self.__setup   = False
        self.ns    = None

    def __str__(self):
        return self.getvalue()

    # - string content methods
    def mangle(self, s):
        '''class/variable name illegalities
        '''
        return TextProtect(s)

    def write(self, s):
        self.content.write(s)

    def writeArray(self, a):
        self.content.write('\n'.join(a))

    def _setContent(self):
        '''override in subclasses.  formats the content in the desired way.
        '''
        raise NotImplementedError, 'abstract method not implemented'

    def getvalue(self):
        if not self.__setup:
            self._setContent()
            self.__setup = True
            
        return self.content.getvalue()

    # - namespace utility methods
    def getNSAlias(self):
        if self.ns:
            return NAD.getAlias(self.ns)
        raise ContainerError, 'no self.ns attr defined in %s' % self.__class__

    def getNSModuleName(self):
        if self.ns:
            return NAD.getModuleName(self.ns)
        raise ContainerError, 'no self.ns attr defined in %s' % self.__class__


# -- containers for services file components

class ServiceContainerBase(ContainerBase):
    def __init__(self):
        ContainerBase.__init__(self)
    
class ServiceHeaderContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.basic = [
            '\nimport urlparse, types',
            'from ZSI.TCcompound import Struct',
            'from ZSI import client',
            'import ZSI'
            ]
        self.types = None
        self.extras = []

    def setTypesModuleName(self, module):
        self.types = module

    def appendImport(self, statement):
        '''append additional import statement(s).
        import_stament -- tuple or list or str 
        '''
        if type(statement) in (list,tuple):
            self.extras += statement
        else:
            self.extras.append(statement)

    def _setContent(self):
        if self.types:
            self.write('from %s import *' % self.types)

        imports = self.basic[:]
        imports += self.extras
        self.writeArray(imports)


class ServiceLocatorContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.serviceName = None
        self.portInfo = []
        self.locatorName = None
        self.portMethods = []

    def setUp(self, service):
        assert isinstance(service, WSDLTools.Service), \
           'expecting WDSLTools.Service instance.'

        self.serviceName = service.name
        for p in service.ports:
            try:
                ab = p.getAddressBinding()
            except WSDLTools.WSDLError, ex:
                self.logger.warning('Skip port(%s), missing address binding' %p.name)
                continue
            if isinstance(ab, WSDLTools.SoapAddressBinding) is False:
                self.logger.warning('Skip port(%s), not a SOAP-1.1 address binding' %p.name)
                continue

            info = (p.getBinding().getPortType().name, p.getBinding().name, ab.location)
            self.portInfo.append(info) 

    def getLocatorName(self):
        '''return class name of generated locator.
        '''
        return self.locatorName

    def getPortMethods(self):
        '''list of get port accessor methods of generated locator class.
        '''
        return self.portMethods

    def _setContent(self):
        if not self.serviceName:
            raise ContainerError, 'no service name defined!'

        self.serviceName = self.mangle(self.serviceName)
        self.locatorName = '%sLocator' %self.serviceName
        locator = ['# Locator', 'class %s:' %self.locatorName, ]
        self.portMethods = []
        for p in self.portInfo:
            ptName = p[0]
            bName  = p[1]
            sAdd   = p[2]
            method = 'get%s' %ptName
            pI = [
                '%s%s_address = "%s"' % (ID1, ptName, sAdd),
                '%sdef get%sAddress(self):' % (ID1, ptName),
                '%sreturn %sLocator.%s_address' % (ID2,
                                                   self.serviceName,ptName),
                '%sdef %s(self, url=None, **kw):' %(ID1, method),
                '%sreturn %sSOAP(url or %sLocator.%s_address, **kw)' \
                % (ID2, bName, self.serviceName, ptName),
                ]
            self.portMethods.append(method)
            locator += pI

        self.writeArray(locator)


class ServiceOperationsClassContainer(ServiceContainerBase):

    def __init__(self, useWSA=False):
        '''Parameters:
        name -- binding name
        property -- resource properties
        useWSA   -- boolean, enable ws-addressing
        name -- binding name
        '''
        ServiceContainerBase.__init__(self)
        self.useWSA = useWSA
        self.rProp = None
        self.bName = None
        self.readerclass = self.writerclass = None
        self.operations = []

    def setReaderClass(self, className):
        '''specify a reader class name, this must be imported
        in service module.
        '''
        self.readerclass = className 

    def setWriterClass(self, className):
        '''specify a writer class name, this must be imported
        in service module.
        '''
        self.writerclass = className 

    def getOperationContainers(self):
        '''retrieve list of ServiceOperationContainer instances.
        '''
        assert self.bName is not None, 'call setUp first.'
        return self.operations

    def addOperation(self, name, port):
        '''
        name -- Operation Name.
        port -- WSDL.Port instance
        '''
        bop = port.getBinding().operations.get(name)
        op = port.getBinding().getPortType().operations.get(name)
        if op is None or bop is None:
            raise Wsdl2PythonError, 'no matching portType/Binding operation(%s)' % bop.name

        c = ServiceOperationContainer(name, useWSA=self.useWSA)
        c.setUp(port)
        self.operations.append(c)

    def setUp(self, port):
        '''This method finds all SOAP Binding Operations, it will skip 
        all bindings that are not SOAP.  
        port -- WSDL.Port instance
        '''
        assert isinstance(port, WSDLTools.Port), 'expecting WSDLTools Port instance'

        self.bName = port.getBinding().name
        self.rProp = port.getBinding().getPortType().getResourceProperties() 

        soap_binding = port.getBinding().findBinding(WSDLTools.SoapBinding)
        if soap_binding is None:
            raise Wsdl2PythonError, 'port(%s) missing WSDLTools.SoapBinding' %port.name

        for bop in port.getBinding().operations:
            soap_bop = bop.findBinding(WSDLTools.SoapOperationBinding)
            if soap_bop is None:
                self.logger.warning(\
                    'Skip port(%s) operation(%s) no SOAP Binding Operation' %(port.name, bop.name),
                )
                continue

            #soapAction = soap_bop.soapAction
            if bop.input is not None:
                soapBodyBind = bop.input.findBinding(WSDLTools.SoapBodyBinding)
                if soapBodyBind is None:
                    self.logger.warning(\
                        'Skip port(%s) operation(%s) Bindings(%s) not supported'\
                        %(port.name, bop.name, bop.extensions)
                    )
                    continue
                
            self.addOperation(bop.name, port)

    def _setContent(self):
        if self.useWSA is True:
            ctorArgs = 'endPointReference=None, **kw'
            epr      = 'self.endPointReference = endPointReference'
        else:
            ctorArgs = '**kw'
            epr      = '# no ws-addressing'

        if self.rProp:
            rprop = 'kw.setdefault("ResourceProperties", ("%s","%s"))'\
                %(self.rProp[0], self.rProp[1])
        else:
            rprop = '# no resource properties'

        methods = [
            '# Methods',
            'class %sSOAP:' % self.bName,
            '%sdef __init__(self, url, %s):' % (ID1, ctorArgs),
            '%skw.setdefault("readerclass", %s)' % (ID2, self.readerclass),
            '%skw.setdefault("writerclass", %s)' % (ID2, self.writerclass),
            '%s%s' % (ID2, rprop),
            '%sself.binding = client.Binding(url=url, **kw)' %ID2,
            '%s%s' % (ID2,epr),
            ]

        for op in self.operations:
            methods += [ op.getvalue() ]

        self.writeArray(methods)


class ServiceOperationContainer(ServiceContainerBase):
    def __init__(self, name, useWSA=False, soapAction=None):
        '''Parameters:
        name -- binding name
        useWSA   -- boolean, enable ws-addressing
        soapAction -- soapaction value for this operation
        '''
        ServiceContainerBase.__init__(self)
        self.name = name
        self.useWSA  = useWSA

        self.port = None
        self.soapaction = None
        self.inputName  = None
        self.outputName = None
        self.inputSimpleType  = None
        self.outputSimpleType = None
        self.inputAction  = None
        self.outputAction = None

    def hasInput(self):
        return self.inputName is not None

    def hasOutput(self):
        return self.outputName is not None

    def isRPC(self):
        binding = self.port.getBinding()
        soap = binding.findBinding(WSDLTools.SoapBinding)
        return soap.style=='rpc'

    def isLiteral(self, input=True):
        binding = self.port.getBinding()
        bop = binding.operations.get(self.name)
        msgrole = bop.input
        if input is False:
            msgrole = bop.output
        body = msgrole.findBinding(WSDLTools.SoapBodyBinding)
        return body.use=='literal'

    def isSimpleType(self, input=True):
        if input is False:
            return self.outputSimpleType
        return self.inputSimpleType

    def getOperation(self):
        return self.port.getBinding().getPortType().get(self.name)

    def getBOperation(self):
        return self.port.getBinding().get(self.name)

    def getOperationName(self):
        return self.name

    def setUp(self, port):
        '''Parameters:
	port -- WSDLTools Port instance.
        '''
        assert isinstance(port, WSDLTools.Port), 'expecting WSDLTools Operation instance'

        self.port = port
        name = self.name
        bop = port.getBinding().operations.get(name)
        op = port.getBinding().getPortType().operations.get(name)
        if op is None or bop is None:
            raise Wsdl2PythonError, 'no matching portType/Binding operation(%s)' %name

        soap_bop = bop.findBinding(WSDLTools.SoapOperationBinding)
        assert soap_bop is not None, 'expecting SOAP Bindings only.'

        # set encodingStyle
        if bop.input is not None:
            sbody = bop.input.findBinding(WSDLTools.SoapBodyBinding)
            self.encodingStyle = None
            if sbody.use == 'encoded':
                assert sbody.encodingStyle == SOAP.ENC,\
                    'Supporting encodingStyle=%s, not %s' %(SOAP.ENC, sbody.encodingStyle)
                self.encodingStyle = sbody.encodingStyle

        self.soapaction=soap_bop.soapAction
        if op.input:
            self.inputName  = op.getInputMessage().name
            self.inputSimpleType = FromMessageGetSimpleElementDeclaration(op.getInputMessage())
            self.inputAction = op.getInputAction()
        if op.output:
            self.outputName = op.getOutputMessage().name
            self.outputSimpleType = FromMessageGetSimpleElementDeclaration(op.getOutputMessage())
            self.outputAction = op.getOutputAction()

    def _setContent(self):
        '''create string representation of operation.
        '''
        kwstring = 'kw = {}'
        tCheck = 'if isinstance(request, %s) is False:' % self.inputName
        bindArgs = ''
        if self.encodingStyle is not None:
            bindArgs = 'encoding="%s", ' %self.encodingStyle

        if self.useWSA:
            wsactionIn = 'wsaction = "%s"' % self.inputAction
            wsactionOut = 'wsaction = "%s"' % self.outputAction
            bindArgs += 'wsaction=wsaction, endPointReference=self.endPointReference, '
            responseArgs = ', wsaction=wsaction'
        else:
            wsactionIn = '# no input wsaction'
            wsactionOut = '# no output wsaction'
            responseArgs = ''

        bindArgs += '**kw)'

        method = [
            '%s# op: %s' % (ID1, self.name),
            '%sdef %s(self, request):' % (ID1, self.name),
            '%s%s' % (ID2, tCheck),
            '%sraise TypeError, "%%s incorrect request type" %% (%s)' %(ID3, 'request.__class__'),
            '%s%s' % (ID2, kwstring),
            '%s%s' % (ID2, wsactionIn),
            '%sself.binding.Send(None, None, request, soapaction="%s", %s'\
            %(ID2, self.soapaction, bindArgs),
        ]
        if self.outputName:
            response = ['%s%s' % (ID2, wsactionOut),]
            response.append('%sresponse = self.binding.Receive(%s.typecode%s)' %(ID2, self.outputName, responseArgs))
            if self.outputSimpleType:
                response.append('%sreturn %s(response)' %(ID2, self.outputName))
            else: 
                rCheck = 'if isinstance(response, %s.typecode.pyclass) is False:' %self.outputName
                response.append('%s%s' %(ID2, rCheck))
                response.append('%sraise TypeError, "%%s incorrect response type" %% (%s)' %(ID3, 'response.__class__'))
                response.append('%sreturn response' %ID2)
            method += response

        self.writeArray(method)


class ServiceMessageContainerBase(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.pName = None
        self.mName = None


class ServiceDocumentLiteralMessageContainer(ServiceMessageContainerBase):
    def __init__(self, simple=None):
        '''
        simple -- primitive type message represents, or None if not primitive.
        '''
        ServiceMessageContainerBase.__init__(self)
        self._simple = simple

    def _setContent(self):
        '''create string representation of doc/lit message container.  If 
        message element is simple(primitive), use python type as base class.
        '''
        message = []
        if self._simple is None:
            message.append('class %s:' %self.mName)
        else:
            message.append('class %s(%s):' %(self.mName, self._simple))

        nsa = self.getNSAlias()
        message.append('%stypecode = %s.%s_Dec()' %(ID1, nsa, self.pName))
        self.writeArray(message)


class ServiceRPCEncodedMessageContainer(ServiceMessageContainerBase):
    '''
    '''
    def __init__(self, input=False):
        '''
        Instance Data: 
           op    -- WSDLTools Operation instance
           bop   -- WSDLTools BindingOperation instance
           input -- boolean input/output
        '''
        ServiceMessageContainerBase.__init__(self)
        self.input = input
        self.op = self.bop = None

    def setUp(self, port, name):
        '''
        port -- WSDLTools.Port instance
        name -- operation name.
        '''
        bop = port.getBinding().operations.get(name)
        op = port.getBinding().getPortType().operations.get(name)

        assert op is not None, 'port has no operation %s' %name
        assert bop is not None, 'port has no binding operation %s' %name

        self.op = op
        self.bop = bop

    def _setContent(self):
        assert self.op is not None and self.bop is not None, 'call setUp.' 

        pname = self.op.name
        msgRole = self.op.input
        msgRoleB = self.bop.input
        if self.input is False:
            pname = '%sResponse' %self.op.name
            msgRole = self.op.output
            msgRoleB = self.bop.output

        sbody = msgRoleB.findBinding(WSDLTools.SoapBodyBinding)
        if not sbody or not sbody.namespace:
            raise WSInteropError, WSISpec.R2717

        assert sbody.use == 'encoded', 'Expecting use=="encoded"'
        encodingStyle = sbody.encodingStyle

        assert encodingStyle == SOAP.ENC,\
            'Supporting encodingStyle=%s, not %s' %(SOAP.ENC, encodingStyle)

        namespace = sbody.namespace
        tcb = MessageTypecodeContainer(\
                  tuple(msgRole.getMessage().parts.data.values()),
              )
        ofwhat = '[%s]' %tcb.getTypecodeList()
        pyclass = msgRole.getMessage().name
        message = [
            'class %s:' %pyclass,
            '%sdef __init__(self):' % ID1,
            ]

        for aname in tcb.getAttributeNames():
            message.append('%sself.%s = None' %(ID2, aname))
        message.append('%sreturn' %ID2)

        message.append('%s.typecode = Struct(pname=("%s","%s"), ofwhat=%s, pyclass=%s, encoded="%s")' \
            %(pyclass, sbody.namespace, pname, ofwhat, pyclass, namespace),)
        self.writeArray(message)


class ServiceRPCLiteralMessageContainer(ServiceMessageContainerBase):
    '''
    '''
    def __init__(self, input=False):
        '''Instance Data: 
               operation -- WSDLTools Operation instance
               boperation -- WSDLTools BindingOperation instance
               input -- boolean input/output
        '''
        ServiceMessageContainerBase.__init__(self)
        self.input = input
        self.op = self.bop = None

    def setUp(self, port, name):
        '''
        port -- WSDLTools.Port instance
        name -- operation name.
        '''
        bop = port.getBinding().operations.get(name)
        op = port.getBinding().getPortType().operations.get(name)

        assert op is not None, 'port has no operation %s' %name
        assert bop is not None, 'port has no binding operation %s' %name

        self.op = op
        self.bop = bop

    def _setContent(self):
        assert self.op is not None, 'call setUp.' 

        operation = self.op
        input = self.input
        pname = operation.name
        msgRole = operation.input
        msgRoleB = self.bop.input
        if input is False:
            pname = '%sResponse' %operation.name
            msgRole = operation.output
            msgRoleB = self.bop.output

        sbody = msgRoleB.findBinding(WSDLTools.SoapBodyBinding)
        if not sbody or not sbody.namespace:            
            raise WSInteropError, WSISpec.R2717

        mName = msgRole.getMessage().name
        parts = msgRole.getMessage().parts.data.values()
        tcb = MessageTypecodeContainer(tuple(parts))
        ofwhat = '[%s]' %tcb.getTypecodeList()
        pyclass = mName

        message = [
            'class %s:' %mName,
            '%sdef __init__(self):' % ID1,
            ]

        anames = tcb.getAttributeNames()
        for aname in anames:
            message.append('%sself.%s = None' %(ID2, aname))
        if not len(anames):
            message.append('%sreturn' %ID2)

        message.append('%s.typecode = Struct(pname=("%s","%s"), ofwhat=%s, pyclass=%s)' \
            %(mName, sbody.namespace, pname, ofwhat, pyclass),)
        self.writeArray(message)


class ServiceEncodedMessageContainer(ServiceMessageContainerBase):

    def _setContent(self):

        message = [
            '# bogus encoded message',
            ]

        self.writeArray(message)


class TypesContainerBase(ContainerBase):
    '''base class for containers of type file components
    '''
    def __init__(self):
        ContainerBase.__init__(self)


class TypesHeaderContainer(TypesContainerBase):
        
    def _setContent(self):
        
        imports = [
            'import ZSI',
            'import ZSI.TCcompound',
            'from ZSI.TC import ElementDeclaration,TypeDefinition',
        ]

        self.writeArray(imports)

class NamespaceClassContainerBase(TypesContainerBase):
    def __init__(self):
        TypesContainerBase.__init__(self)
        

class NamespaceClassHeaderContainer(NamespaceClassContainerBase):

    def _setContent(self):

        head = [
            '#' * 30,
            '# targetNamespace',
            '# %s' % self.ns,
            '#' * 30 + '\n',
            'class %s:' % self.getNSAlias(),
            '%stargetNamespace = "%s"' % (ID1, self.ns)
            ]

        self.writeArray(head)

class NamespaceClassFooterContainer(NamespaceClassContainerBase):

    def _setContent(self):

        foot = [
            '# end class %s (tns: %s)' % (self.getNSAlias(), self.ns),
            ]

        self.writeArray(foot)


BTI = BaseTypeInterpreter()
class TypecodeContainerBase(TypesContainerBase):
    '''Base class for all classes representing anything
    with element content.

    class variables:
        mixed_content_aname -- text content will be placed in this attribute.
        attributes_aname -- attributes will be placed in this attribute.
        func_aname -- function that takes name, and returns aname.
   
    '''
    mixed_content_aname = 'text'
    attributes_aname = 'attrs'
    func_aname = TextProtectAttributeName
    func_aname = staticmethod(func_aname)

    suffix_def = lambda n: '%s_Def'
    suffix_dec =  lambda n: '%s_Dec'
    
    def __init__(self):
        TypesContainerBase.__init__(self)
        self.name = None

        # attrs for model groups and others with elements, tclists, etc...
        self.allOptional = False
        self.mgContent = None
        self.contentFlattened = False
        self.elementAttrs = []
        self.elementsSet = False
        self.tcListElements = []
        self.tcListSet = False

        self.localTypes = []

        # used when processing nested anonymous types
        self.parentClass = None

        # used when processing attribute content

        self.mixed = False
        self.extraFlags = ''

        self.attributeContent = None
        self.attrComponents = []

    def getvalue(self):
        out = ContainerBase.getvalue(self)
        for item in self.localTypes:
            content = None
            assert True is item.isElement() is item.isLocal(), 'expecting local elements only'

            etp = item.content
            qName = item.getAttribute('type')
            if not qName:
                etp = item.content
                local = True
            else:
                etp = item.getTypeDefinition('type')

            if etp is None:
                if local is True:
                    content = ElementLocalComplexTypeContainer()
                else:
                    content = ElementSimpleTypeContainer()
            elif etp.isLocal() is False:
                content = ElementGlobalDefContainer()
            elif etp.isSimple() is True:
                content = ElementLocalSimpleTypeContainer()
            elif etp.isComplex():
                content = ElementLocalComplexTypeContainer()
            else:
                raise Wsdl2PythonError, "Unknown element declaration: %s" %item.getItemTrace()

            content.setUp(item)

            out += '\n\n'
            if self.parentClass:
                content.parentClass = \
                    '%s.%s' %(self.parentClass, self.getClassName())
            else:
                content.parentClass = '%s.%s' %(self.getNSAlias(), self.getClassName())

            for l in content.getvalue().split('\n'):
                if l: out += '%s%s\n' % (ID1, l)
                else: out += '\n'

            out += '\n\n'

        return out

    def getAttributeName(self, name):
        '''represents the aname
        '''
        if self.func_aname is None:
            return name
        assert callable(self.func_aname), \
            'expecting callable method for attribute func_aname, not %s' %type(self.func_aname)
        f = self.func_aname
        return f(name)

    def getMixedTextAName(self):
        '''returns an aname representing mixed text content.
        '''
        return self.getAttributeName(self.mixed_content_aname)

    def getClassName(self):

        if not self.name:
            raise ContainerError, 'self.name not defined!'
        if not hasattr(self.__class__, 'type'):
            raise ContainerError, 'container type not defined!'

        suffix = self.__class__.type

        #if self.__class__.type == 'def':
        #    suffix = '_Def'
        #elif self.__class__.type == 'dec':
        #    suffix = '_Dec'
        #else:
        #    raise ContainerError, 'unknown container instance type: %s' \
        #          % self.__class__.type
        
        return self.mangle( self.name + suffix )


    def getPyClass(self):
        '''Name of generated inner class that will be specified as pyclass.
        '''
        return 'Holder'

    def getPyClassDefinition(self):
        '''Return a list containing pyclass definition.
        '''
        definition = []
        definition.append('%sclass %s:' %(ID3,self.getPyClass()))
        definition.append('%stypecode = self' %ID4)
        definition.append('%sdef __init__(self):' %ID4)
        definition.append('%s# pyclass' %ID5)

        # JRB HACK need to call _setElements via getElements
        self.getElements()
        # JRB HACK need to indent additional one level
        for el in self.elementAttrs:
            definition.append('%s%s' %(ID2,el))
        definition.append('%sreturn' %ID5)

        return definition

    def nsuriLogic(self):
        '''set a variable "ns" that represents the targetNamespace in
        which this item is defined.  Used for namespacing local elements.
        '''
        if self.parentClass:
            return 'ns = %s.%s.schema' %(self.parentClass, self.getClassName())
        return 'ns = %s.%s.schema' %(self.getNSAlias(), self.getClassName())

    def schemaTag(self):
        if self.ns is not None:
            return 'schema = "%s"' % self.ns
        raise ContainerError, 'failed to set schema targetNamespace(%s)' %(self.__class__)
    
    def typeTag(self):
        if self.name is not None:
            return 'type = (schema, "%s")' % self.name
        raise ContainerError, 'failed to set type name(%s)' %(self.__class__)
    
    def literalTag(self):
        if self.name is not None:
            return 'literal = "%s"' % self.name
        raise ContainerError, 'failed to set element name(%s)' %(self.__class__)

    def getExtraFlags(self):
        if self.mixed:
            self.extraFlags += 'mixed=True, mixed_aname="%s", ' %self.getMixedTextAName()

        return self.extraFlags

    def simpleConstructor(self, superclass=None):

        if superclass:
            return '%s.__init__(self, **kw)' % superclass
        else:
            return 'def __init__(self, **kw):'

    def pnameConstructor(self, superclass=None):

        if superclass:
            return '%s.__init__(self, pname, **kw)' % superclass
        else:
            return 'def __init__(self, pname, **kw):'

    # _flattenContent() is a pre-processor called before either the
    # elements or typecode strings are generated.  it flattens out
    # embedded/anonymous model group elements.  _setElements() and
    # _setTypecodeList() skip the model groups and pay attention
    # to the elements contained therein.

    def _flattenContent(self):
        if not self.contentFlattened:
            self.contentFlattened = True
            if type(self.mgContent) is not tuple:
                raise TypeError, 'expecting tuple for mgContent, not: %s' %type(self.mgContent)
            for c in self.mgContent:
                if c.isChoice() or c.isSequence():
                    self.mgContent += c.content

    # getElements() and _setElements() generate the class variable
    # elements lists in definition classes

    def _setElements(self):

        if type(self.mgContent) is not tuple:
            raise TypeError, 'expecting tuple for mgContent got %s' %type(self.mgContent)

        for c in self.mgContent:
            if c.isDeclaration() and c.isElement():
                if c.getAttribute('name'):
                    e = '%sself.%s = None' \
                        %(ID3, self.getAttributeName(c.getAttribute('name')))
                    self.elementAttrs.append(e)
                elif c.isWildCard():
                    e = '%sself.%s = None' % (ID3, self.getAttributeName(c.getAttribute('name')))
                else:
                    raise ContainerError, 'This is not an element declaration: %s' %c.getItemTrace()
            elif c.isReference():
                if c.getAttribute('ref')[1]:
                    e = '%sself._%s = None' \
                        % (ID3,self.mangle(c.getAttribute('ref')[1]))
                    self.elementAttrs.append(e)
                else:
                    raise ContainerError, 'ouch as well'
            elif c.isSequence() or c.isChoice():
                if not self.contentFlattened:
                    raise ContainerError, 'unflattened content found!'
                else:
                    # model group contents already flattened
                    pass
            else:
                raise ContainerError, 'unexpected item: %s' % c.getItemTrace()

    def getElements(self):
        if not self.elementsSet:
            self._flattenContent()
            self._setElements()
            self.elementsSet = True

        return '\n'.join(self.elementAttrs)

    def _setTypecodeList(self):
        '''getTypecodeList and _setTypecodeList generates the TClist 
        strings where appropriate.

            mgContent -- 
            localTypes -- produce local class definitions later
            tcListElements -- elements, local/global 
 
        '''
        if type(self.mgContent) is not tuple:
            raise TypeError, 'expecting tuple for mgContent got %s' %type(self.mgContent)

        for c in self.mgContent:
            tc = TcListComponentContainer()
            min,max,nil = self._getOccurs(c)
            tc.setOccurs(min, max, nil)

            if c.isDeclaration() and c.isElement():
                global_type = c.getAttribute('type')
                content = getattr(c, 'content', None)
                if c.isLocal() and c.isQualified() is False:
                    tc.unQualified()

                if c.isWildCard():
                    tc.setStyleAnyElement()
                elif global_type is not None:
                    tc.name = c.getAttribute('name')
                    if self._isXSD(c):
                        tpc = BTI.get_typeclass(c.getAttribute('type')[1],
                                                     c.getAttribute('type')[0])
                        tc.klass = tpc
                    elif self._isRecursiveElement(c):
                        tc.setStyleRecursion()
                    else:
                        ns = c.getAttribute('type')[0]
                        tc.klass = '%s.%s_Def' % (NAD.getAlias(ns),
                                                  c.getAttribute('type')[1])
                elif content is not None and content.isLocal() and content.isComplex():
                    tc.name = c.getAttribute('name')
                    tc.klass = 'self.__class__.%s_Dec' % tc.name
                    tc.setStyleElementReference()
                    self.localTypes.append(c)
                elif content is not None and content.isLocal() and content.isSimple():
                    # Local Simple Type
                    tc.name = c.getAttribute('name')
                    tc.klass = 'self.__class__.%s_Dec' % tc.name
                    tc.setStyleElementReference()
                    self.localTypes.append(c)
                else:
                    raise ContainerError, 'unexpected item: %s' % c.getItemTrace()

            elif c.isReference():
                # element references
                ns = c.getAttribute('ref')[0]
                tc.klass = '%s.%s_Dec' % (NAD.getAlias(ns),
                                          c.getAttribute('ref')[1])
                tc.setStyleElementReference()
            elif c.isSequence() or c.isChoice():
                if not self.contentFlattened:
                    raise ContainerError, 'unflattened content found!'
                continue
            else:
                raise ContainerError, 'unexpected item: %s' % c.getItemTrace()

            self.tcListElements.append(tc)

    def getTypecodeList(self):
        if not self.tcListSet:
            self._flattenContent()
            self._setTypecodeList()
            self.tcListSet = True

        list = []
        for e in self.tcListElements:
            list.append(str(e))

        return ', '.join(list)

    # the following _methods() are utility methods used during
    # TCList generation, et al.
    
    def _isXSD(self, c):
        if c.getAttribute('type'):
            if c.getAttribute('type')[0] in SCHEMA.XSD_LIST:
                return True
            
        return False

    def _isRecursiveElement(self, c):
        if self.name == c.getAttribute('type')[1]:
            if self.ns == c.getAttribute('type')[0]:
                return True

        return False

    def _getOccurs(self, e):
        
        nillable = e.getAttribute('nillable')
        if nillable == 'true':
            nillable = True
        else:
            nillable = False
            
        maxOccurs = e.getAttribute('maxOccurs')
        if maxOccurs == 'unbounded':
            maxOccurs = '"%s"' %maxOccurs
            
        minOccurs = e.getAttribute('minOccurs')
        
        if self.allOptional is True:
            #JRB Hack
            minOccurs = '0'
            maxOccurs = '"unbounded"'

        return minOccurs,maxOccurs,nillable

    # these methods are used when generating attribute content:

    def _setAttributes(self):
        '''set up generated typecode.
        '''
        if not self.attributeContent:
            return
        self.attrComponents.append('# attribute handling code')
        self.attrComponents.append('self.attribute_typecode_dict.update(%s.%s.attribute_typecode_dict)' \
            %(self.getNSAlias(), self.getClassName()))

        for a in self.attributeContent:
            if a.isWildCard() and a.isAttribute() and a.isDeclaration():
                self.attrComponents.append('self.attribute_typecode_dict[("%s","anyAttribute")] = ZSI.TC.AnyElement()' % (SCHEMA.XSD3))
            elif a.isDeclaration():
                t = a.getAttribute('type')
                try:
                    tc = BTI.get_typeclass(t[1], t[0])
                except:
                    raise ContainerError, 'type attribute(%s) not primitive type?: %s' \
                        %(t, a.getItemTrace())

                key = None
                if a.getAttribute('form') == 'qualified':
                    key = '("%s","%s")' % ( a.getTargetNamespace(),
                                            a.getAttribute('name') )
                elif a.getAttribute('form') == 'unqualified':
                    key = '"%s"' % a.getAttribute('name')
                else:
                    raise ContainerError, \
                          'attribute form must be un/qualified %s' \
                          % a.getAttribute('form')
                self.attrComponents.append('self.attribute_typecode_dict[%s] = %s()' % (key, tc))
                
            elif a.isReference():
                if a.isAttributeGroup():
                    # flatten 'em out....
                    for ga in a.getAttributeGroup().getAttributeContent():
                        if not ga.isAttributeGroup():
                            self.attributeContent += (ga,)
                    continue
                ga = a.getAttributeDeclaration()
                tp = ga.getTypeDefinition('type')
                if tp is None:
                    # built in simple type
                    try:
                        namespace,typeName = ga.getAttribute('type')
                    except TypeError, ex:
                        raise ContainerError,\
                              'no type in attribute declaration!'
                    tc = BTI.get_typeclass(typeName, namespace)
                    key = '("%s","%s")' \
                          % (ga.getTargetNamespace(),ga.getAttribute('name'))
                    self.attrComponents.append('self.attribute_typecode_dict[%s] = %s()' % (key, tc))
                else:
                    typeName = tp.getAttribute('name')
                    namespace = tp.getTargetNamespace()
                    alias = NAD.getAlias(namespace)
                    key = '("%s","%s")' \
                          % (ga.getTargetNamespace(),ga.getAttribute('name'))
                    self.attrComponents.append('self.attribute_typecode_dict[%s] = %s.%s_Def(None)' % (key, alias, typeName))
                
            else:
                raise ContainerError, 'unknown attribute type!'


    def getAttributes(self):
        return self.attrComponents

    def getBasesLogic(self, indent):
        try:
            prefix = NAD.getAlias(self.sKlassNS)
        except WsdlGeneratorError, ex:
            # XSD or SOAP
            raise

        bases = []
        bases.append(\
            'if %s.%s_Def not in %s.%s.__bases__:'\
            %(NAD.getAlias(self.sKlassNS), self.sKlass, self.getNSAlias(), self.getClassName()),
        )
        bases.append(\
            '%sbases = list(%s.%s.__bases__)'\
            %(ID1,self.getNSAlias(),self.getClassName()),
        )
        bases.append(\
            '%sbases.insert(0, %s.%s_Def)'\
            %(ID1,NAD.getAlias(self.sKlassNS), self.sKlass),
        )
        bases.append(\
            '%s%s.%s.__bases__ = tuple(bases)'\
            %(ID1, self.getNSAlias(), self.getClassName())
        )

        s = ''
        for b in bases:
            s += '%s%s\n' % (indent, b)

        return s


class MessageTypecodeContainer(TypecodeContainerBase):
    '''Used for RPC style messages, where we have 
    serveral parts serialized within a rpc wrapper name.
    '''
    def __init__(self, parts=None):
        TypecodeContainerBase.__init__(self)
        self.mgContent = parts

    def _getOccurs(self, e):
        '''return a 3 item tuple 
        '''
        minOccurs = maxOccurs = '1'
        nillable = True
        return minOccurs,maxOccurs,nillable

    def _setTypecodeList(self):
        assert type(self.mgContent) is tuple,\
            'expecting tuple for mgContent not: %s' %type(self.mgContent)

        for p in self.mgContent:
            # JRB
            #   not sure what needs to be set for tc, this should be
            #   the job of the constructor or a setUp method.
            min,max,nil = self._getOccurs(p)
            if p.element:
                raise  WSInteropError, WSISpec.R2203
            elif p.type: 
                nsuri,name = p.type
                tc = RPCMessageTcListComponentContainer(qualified=False)
                tc.setOccurs(min, max, nil)
                tc.name = p.name
                if nsuri in SCHEMA.XSD_LIST:
                    tpc = BTI.get_typeclass(name, nsuri)
                    tc.klass = tpc
                else:
                    tc.klass = '%s.%s_Def' % (NAD.getAlias(nsuri), name)
            else:
                raise ContainerError, 'part must define an element or type attribute'

            self.tcListElements.append(tc)

    def getTypecodeList(self):
        if not self.tcListSet:
            self._setTypecodeList()
            self.tcListSet = True

        list = []
        for e in self.tcListElements:
            list.append(str(e))
        return ', '.join(list)

    def getAttributeNames(self):
        '''returns a list of anames representing the parts
        of the message.
        '''
        return map(lambda e: self.getAttributeName(e.name), self.tcListElements)

    def setParts(self, parts):
        self.mgContent = parts


class TcListComponentContainer(TypecodeContainerBase):
    '''Encapsulates a single value in the TClist list.
    it inherits TypecodeContainerBase only to get the mangle() method,
    it does not call the baseclass ctor.
    '''
    
    def __init__(self, qualified=True):
        '''
        qualified -- qualify element.  All GEDs should be qualified,
            but local element declarations qualified if form attribute
            is qualified, else they are unqualified. Only relevant for
            standard style.
        '''
        self.qualified = qualified
        self.name = None
        self.klass = None
        
        self.min = None
        self.max = None
        self.nil = None
        self.style = None
        self.setStyleElementDeclaration()

    def setOccurs(self, min, max, nil):
        self.min = min
        self.max = max
        self.nil = nil

    def setStyleElementDeclaration(self):
        '''set the element style.
            standard -- GED or local element
        '''
        self.style = 'standard'

    def setStyleElementReference(self):
        '''set the element style.
            ref -- element reference
        '''
        self.style = 'ref'

    def setStyleAnyElement(self):
        '''set the element style.
            anyElement -- <any> element wildcard
        '''
        self.name = 'any'
        self.style = 'anyElement'

    def setStyleRecursion(self):
        '''set the element style.
            recursion -- do lazy evaluation of element/type because
               the element contains an instance of itself.
        '''
        self.style = 'recursion'

    def unQualified(self):
        '''Do not qualify element.
        '''
        self.qualified = False

    def _getOccurs(self):
        return 'minOccurs=%s, maxOccurs=%s, nillable=%s' \
               % (self.min, self.max, self.nil)

    def _getvalue(self):
        if self.style == 'standard':
            pname = '"%s"' %self.name
            if self.qualified is True:
                pname = '(ns,"%s")' %self.name
            return '%s(pname=%s, aname="%s", %s, encoded=kw.get("encoded"))' \
                   % (self.klass, pname,
                      self.getAttributeName(self.name), self._getOccurs())
        elif self.style == 'ref':
            return '%s(%s, encoded=kw.get("encoded"))' % (self.klass, self._getOccurs())
        elif self.style == 'anyElement':
            return 'ZSI.TC.AnyElement(aname="%s", %s)' \
                %(self.getAttributeName(self.name), self._getOccurs())
        elif self.style == 'recursion':
            return 'ZSI.TC.AnyElement(aname="%s", %s)' \
                % (self.getAttributeName(self.name), self._getOccurs())

    def __str__(self):
        return self._getvalue()
   
 
class RPCMessageTcListComponentContainer(TcListComponentContainer):
    '''Container for rpc/literal rpc/encoded message typecode.
    '''
    def __init__(self, qualified=True, encoded=None):
        '''
        encoded -- encoded namespaceURI, if None treat as rpc/literal.
        '''
        TcListComponentContainer.__init__(self, qualified=qualified)
        self._encoded = encoded
 
    def _getvalue(self):
        encoded = self._encoded
        if encoded is not None:
            encoded = '"%s"' %self._encoded

        if self.style == 'standard':
            pname = '"%s"' %self.name
            if self.qualified is True:
                pname = '(ns,"%s")' %self.name
            return '%s(pname=%s, aname="%s", encoded=%s, %s)' \
                   %(self.klass, pname, self.getAttributeName(self.name), 
                     encoded, self._getOccurs())
        elif self.style == 'ref':
            return '%s(encoded=%s, %s)' % (self.klass, encoded, self._getOccurs())
        elif self.style == 'anyElement':
            return 'ZSI.TC.AnyElement(aname="%s", %s)' \
                %(self.getAttributeName(self.name), self._getOccurs())
        elif self.style == 'recursion':
            return 'ZSI.TC.AnyElement(aname="%s", %s)' \
                % (self.getAttributeName(self.name), self._getOccurs())


class ElementSimpleTypeContainer(TypecodeContainerBase):

    type = DEC

    def setUp(self, tp):
        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        qName = tp.getAttribute('type')
        self.sKlass = BTI.get_typeclass(qName[1], qName[0])

    def _setContent(self):

        element = [
            '%sclass %s(%s, ElementDeclaration):' % (ID1, self.getClassName(),
                                                     self.sKlass),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "%s"' % (ID3, self.getAttributeName(self.name)),
            '%s%s' % (ID3, self.simpleConstructor(self.sKlass)),
            ]

        self.writeArray(element)


class ElementLocalSimpleTypeContainer(TypecodeContainerBase):
    '''local simpleType container
    '''
    type = DEC 
    def _setContent(self):

        element = [
            '%sclass %s(%s, ElementDeclaration):' % (ID1, self.getClassName(),
                                                     self.sKlass),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "%s"' % (ID3, self.getAttributeName(self.name)),
            '%s%s' % (ID3, self.simpleConstructor(self.sKlass)),
            ]

        self.writeArray(element)

    def setUp(self, tp):
        assert tp.isElement() is True and tp.content is not None and \
            tp.content.isLocal() is True and tp.content.isSimple() is True ,\
            'expecting local simple type: %s' %tp.getItemTrace()

        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        content = tp.content.content
        if content.isRestriction():
            try:
                 base = content.getTypeDefinition()
            except SchemaError, ex:
                 base = None

            qName = content.getAttributeBase()
            if base is None:
                self.sKlass = BTI.get_typeclass(qName[1], qName[0])
                return

            raise Wsdl2PythonError, 'unsupported local simpleType restriction: %s' \
                %tp.content.getItemTrace()

        if content.isList():
            try:
                 base = content.getTypeDefinition()
            except SchemaError, ex:
                 base = None

            if base is None:
                qName = content.getItemType()
                self.sKlass = BTI.get_typeclass(qName[1], qName[0])
                return

            raise Wsdl2PythonError, 'unsupported local simpleType List: %s' \
                %tp.content.getItemTrace()

        if content.isUnion():
            raise Wsdl2PythonError, 'unsupported local simpleType Union: %s' \
                %tp.content.getItemTrace()

        raise Wsdl2PythonError, 'unexpected schema item: %s' \
            %tp.content.getItemTrace()


class ElementLocalComplexTypeContainer(TypecodeContainerBase):

    type = DEC

    def setUp(self, tp):
        '''
        {'xsd':['annotation', 'simpleContent', 'complexContent',\
        'group', 'all', 'choice', 'sequence', 'attribute', 'attributeGroup',\
        'anyAttribute', 'any']}
        '''
        # JRB HACK SUPPORTING element/no content.
        assert tp.isElement() is True and \
            (tp.content is None or (tp.content.isComplex() is True and tp.content.isLocal() is True)),\
            'expecting element w/local complexType not: %s' %tp.content.getItemTrace()

        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()

        complex = tp.content
        # JRB HACK SUPPORTING element/no content.
        if complex is None:
            self.mgContent = ()
            return

        self.mgContent = None
        if complex.content is None:
            self.mgContent = ()
        elif complex.content.isSimple() is True:
            if complex.content.content.isExtension() is True:
                # XXX HACK 
                #self.mgContent = complex.content.content
                self.mgContent = ()
                return
            elif complex.content.content.isRestriction() is True:
                # XXX HACK 
                #self.mgContent = complex.content.content
                self.mgContent = ()
                return
            raise ContainerError,\
               'not implemented local complexType/simpleContent: %s' %tp.getItemTrace()
        elif complex.content.isComplex() is True:
            if complex.content.content is None:
                self.mgContent = ()
            elif complex.content.content.isExtension() is True and\
                 complex.content.content.content is not None and\
                 complex.content.content.content.isModelGroup() is True:
                self.mgContent = complex.content.content.content.content
            elif complex.content.content.isRestriction() is True and\
                 complex.content.content.content is not None and\
                 complex.content.content.content.isModelGroup() is True:
                self.mgContent = complex.content.content.content.content
        elif complex.content.isModelGroup() is True:
            self.mgContent = complex.content.content

        if self.mgContent is None: 
            self.mgContent = ()
        assert type(self.mgContent) is tuple, 'XXX: %s' %self.mgContent.getItemTrace()

    def _setContent(self):

        element = [
            '%sclass %s(ZSI.TCcompound.Struct, ElementDeclaration):' \
            % (ID1,self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            #'%s' % self.getElements(),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "%s"' % (ID3, self.getAttributeName(self.name)),
            ]

        # Write out a pyclass
        for l in self.getPyClassDefinition():
            element.append(l)

        element.append(\
            '%sZSI.TCcompound.Struct.__init__(self, %s, TClist, inorder=0, **kw)' \
            %(ID3, self.getPyClass())
        )
        self.writeArray(element)

class ElementGlobalDefContainer(TypecodeContainerBase):

    type = DEC

    def setUp(self, element):
        self.name = element.getAttribute('name')
        self.ns = element.getTargetNamespace()

        tp = element.getTypeDefinition('type')
        self.sKlass = tp.getAttribute('name')
        self.sKlassNS = tp.getTargetNamespace()


    def _setContent(self):
        '''GED defines element name, so also define typecode aname
        '''
        element = [
            '%sclass %s(ElementDeclaration):' % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "%s"' % (ID3, self.getAttributeName(self.name)),
            '%s'   % self.getBasesLogic(ID3),
            '%s%s.%s_Def.__init__(self, **kw)' \
            % (ID3, NAD.getAlias(self.sKlassNS), self.sKlass),
            ]

        self.writeArray(element)


class DerivedTypeDefContainer(TypecodeContainerBase):
    '''JRB Only used for ComplexContent??
    '''
    type = DEF

    def __init__(self):
        '''
        restriction
        extension
        extType -- used in figuring attrs for extensions
        '''
        TypecodeContainerBase.__init__(self)
        self.restriction = False
        self.extension = False
        self.extType = None
        self._is_array = False
        self._kw_array = None

    def setUp(self, tp):
        '''complexContent/[extension,restriction]
        '''
        assert tp.content.isComplex() is True and \
            (tp.content.content.isRestriction() or tp.content.content.isExtension() is True),\
            'expecting complexContent/[extension,restriction]'

        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        derivation = tp.content.content
        try:
            base = derivation.getTypeDefinition('base')
        except XMLSchema.SchemaError, ex:
            base = None

        if base is None:
            # SOAP-ENC:Array expecting arrayType attribute reference
            qname = derivation.getAttributeQName('base')
            if qname is None:
                raise ContainerError, 'No arrayType declared for Array: %s'\
                    %derivation.getItemTrace()

            self.sKlass = BTI.get_typeclass(qname[1], qname[0])
            self.sKlassNS = qname[0]

            self._is_array = True
            self._kw_array = {'atype':None, 'id3':ID3, 'ofwhat':None}
            attr = None
            for a in derivation.getAttributeContent():
                assert a.isAttribute() is True,\
                    'only attribute content expected: %s' %a.getItemTrace()

                if a.isReference() is True:
                    if a.getAttribute('ref') == (SOAP.ENC,'arrayType'):
                        self._kw_array['atype'] = a.getAttributeQName((WSDL.BASE, 'arrayType'))
                        attr = a
                        break

            qname = self._kw_array.get('atype')
            if attr is not None:
                qname = self._kw_array.get('atype')
                ncname = qname[1].strip('[]')
                namespace = qname[0]
                try:
                    ofwhat = attr.getSchemaItem(XMLSchema.TYPES, namespace, ncname)
                except XMLSchema.SchemaError, ex:
                    ofwhat = None

                if ofwhat is None:
                    self._kw_array['ofwhat'] = BTI.get_typeclass(ncname, namespace)
                else:
                    self._kw_array['ofwhat'] = GetClassNameFromSchemaItem(ofwhat)

                if self._kw_array['ofwhat'] is None:
                    raise ContainerError, 'For Array could not resolve ofwhat typecode(%s,%s): %s'\
                        %(namespace, ncname, derivation.getItemTrace())
       
        else:
            self.sKlass = base.getAttribute('name')
            self.sKlassNS = base.getTargetNamespace()

        if derivation.isRestriction():
            self.restriction = True
            self.extension = False
            if hasattr(tp, 'getAttributeContent'):
                self.attributeContent = tp.getAttributeContent()
        elif derivation.isExtension():
            self.restriction = False
            self.extension = True
            attrs = []
            if hasattr(tp, 'getAttributeContent') and \
                tp.getAttributeContent() != None:
                attrs += tp.getAttributeContent()
            if hasattr(derivation, 'getAttributeContent') and \
                derivation.getAttributeContent():
                attrs += derivation.getAttributeContent()
            if attrs:
                self.attributeContent = attrs
                self.extType = derivation
        else:
            raise Wsdl2PythonError, 'neither restriction nor extension?'

        if derivation.content is not None \
            and derivation.content.isModelGroup():
            group = derivation.content
            if group.isReference():
                group = group.getModelGroupReference()
            self.mgContent = group.content
        elif derivation.content:
            raise Wsdl2PythonError, \
                'expecting model group, not: %s' %derivation.content.getItemTrace()
        else:
            self.mgContent = ()

    def _setContent(self):
        '''JRB What is the difference between instance data
        ns, name, -- type definition?
        sKlass, sKlassNS? -- element declaration?
        '''
        definition = []
        if self._is_array is True:
            # 
            # SOAP-ENC:Array
            if _is_xsd_or_soap_ns(self.sKlassNS) is False and self.sKlass == 'Array':
                raise ContainerError, 'unknown type: (%s,%s)'\
                    %(self.sKlass, self.sKlassNS)
            definition.append(\
                '%sclass %s(ZSI.TC.Array, TypeDefinition):' % (ID1, self.getClassName()),
            )
            definition.append(\
                '%s%s' % (ID2, self.schemaTag()),
            )
            definition.append(\
                '%s%s' % (ID2, self.typeTag()),
            )
            definition.append(\
                '%sattribute_typecode_dict = {}' % ID2,
            )
            definition.append(\
                '%s%s' % (ID2, self.pnameConstructor()),
            )
            # No need to xsi:type array items since specify with
            # SOAP-ENC:arrayType attribute.
            definition.append(\
                '%(id3)sofwhat = %(ofwhat)s(None, typed=False)' %self._kw_array,
            )
            definition.append(\
                '%(id3)satype = %(atype)s' %self._kw_array,
            )
            definition.append(\
                '%s%s.__init__(self, atype, ofwhat, pname=pname, childnames=\'item\', **kw)'
                %(ID3, self.sKlass),
            )
        else:
            definition.append(\
                '%sclass %s(TypeDefinition):' % (ID1, self.getClassName()),
            )
            definition.append(\
                '%s%s' % (ID2, self.schemaTag()),
            )
            definition.append(\
                '%s%s' % (ID2, self.typeTag()),
            )
            definition.append(\
                '%sattribute_typecode_dict = {}' % ID2,
            )
            definition.append(\
                '%s%s' % (ID2, self.pnameConstructor()),
            )
            definition.append(\
                '%s%s' % (ID3, self.nsuriLogic()),
            )
            definition.append(\
                '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            )
            definition.append(\
                '%s'   % self.getBasesLogic(ID3),
            )
            definition.append(\
                '%s%s.%s_Def.__init__(self, pname, **kw)' \
                %(ID3, NAD.getAlias(self.sKlassNS), self.sKlass),
            )
            if self.restriction:
                definition.append(\
                    '%sself.ofwhat = tuple(TClist)' %ID3,
                ) 
            else:
                definition.append(\
                    '%sself.ofwhat += tuple(TClist)' %ID3,
                ) 

        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))

        self.writeArray(definition)


class ModelGroupContainer(TypecodeContainerBase):

    type = DEF

    def setUp(self, tp, empty=False):
        '''Problematic, loose all model group information.
        <all>, <choice>, <sequence> ..

           tp -- type definition
           empty -- no model group, just use as a dummy holder.
        '''
        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        self.mixed = tp.isMixed()
        self.attributeContent = tp.getAttributeContent()

        self.mgContent = ()
        if empty is True or tp.content is None:
            pass

        elif tp.content.isReference() is True: 
            ref = tp.content.getModelGroupReference()
            if ref.content.content is not None:
               self.mgContent = ref.content.content
        elif tp.content.content is not None:
            self.mgContent = tp.content.content

    def _setContent(self):

        definition = [
            '%sclass %s(ZSI.TCcompound.Struct, TypeDefinition):'
            % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            #'%s'   % self.getElements(),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            ]

        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))

        # write out a pyclass
        for l in self.getPyClassDefinition():
            definition.append(l)

        definition.append(\
            '%sZSI.TCcompound.Struct.__init__(self, %s, TClist, pname=pname, inorder=0, %s**kw)' \
            %(ID3, self.getPyClass(), self.getExtraFlags())
        )

        self.writeArray(definition)


class SimpleTypeContainer(TypecodeContainerBase):
    type = DEF

    def __init__(self):
        '''
        Instance Data From TypecodeContainerBase NOT USED...
           mgContent
        '''
        TypecodeContainerBase.__init__(self)

    def setUp(self, tp):
        raise NotImplementedError, 'abstract method not implemented'

    def _setContent(self, tp):
        raise NotImplementedError, 'abstract method not implemented'

    def getPythonType(self):
        pyclass = eval(str(self.sKlass))
        if issubclass(pyclass, ZSI.TC.String):
            return 'str'
        if issubclass(pyclass, ZSI.TC.Ilong) or issubclass(pyclass, ZSI.TC.IunsignedLong):
            return 'long'
        if issubclass(pyclass, ZSI.TC.Boolean) or issubclass(pyclass, ZSI.TC.Integer):
            return 'int'
        if issubclass(pyclass, ZSI.TC.Decimal):
            return 'float'
        if issubclass(pyclass, ZSI.TC.Gregorian) or issubclass(pyclass, ZSI.TC.Duration):
            return 'tuple'
        return None

    def getPyClassDefinition(self):
        definition = []
        pt = self.getPythonType()
        if pt is not None:
            definition.append('%sclass %s(%s):' %(ID3,self.getPyClass(),pt))
            definition.append('%stypecode = self' %ID4)
        return definition


class RestrictionContainer(SimpleTypeContainer):
    '''
       simpleType/restriction
    '''

    def setUp(self, tp):
        assert tp.isSimple() is True and tp.isDefinition() is True and \
            tp.content.isRestriction() is True,\
            'expecting simpleType restriction, not: %s' %tp.getItemTrace()

        if tp.content is None:
            raise Wsdl2PythonError, \
                  'empty simpleType defintion: %s' %tp.getItemTrace()

        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        self.sKlass = None

        base = tp.content.getAttribute('base')
        if base is not None:
            try:
                item = tp.content.getTypeDefinition('base')
            except SchemaError, ex:
                pass

            if item is None:
                self.sKlass = BTI.get_typeclass(base[1], base[0])
                return

            raise Wsdl2PythonError, \
                'Not Supporting simpleType/Restriction w/User-Defined Base: ' %item.getItemTrace()

        sc = tp.content.getSimpleTypeContent()
        if sc is not None and True is sc.isSimple() is sc.isLocal() is sc.isDefinition():
            base = None
            if sc.content.isRestriction() is True:
                try:
                    item = tp.content.getTypeDefinition('base')
                except SchemaError, ex:
                    pass

                if item is None:
                    base = sc.content.getAttribute('base')
                    if base is not None:
                        self.sKlass = BTI.get_typeclass(base[1], base[0])
                        return
                    raise Wsdl2PythonError, \
                        'Not Supporting simpleType/Restriction w/User-Defined Base: '\
                        %item.getItemTrace()

                raise Wsdl2PythonError, \
                    'Not Supporting simpleType/Restriction w/User-Defined Base: '\
                    %item.getItemTrace()

            if sc.content.isList() is True:
                raise Wsdl2PythonError, \
                      'iction base in subtypes: %s'\
                      %sc.getItemTrace()

            if sc.content.isUnion() is True:
                raise Wsdl2PythonError, \
                      'could not get restriction base in subtypes: %s'\
                      %sc.getItemTrace()

            return

        raise Wsdl2PythonError, 'No Restriction @base/simpleType: %s' %tp.getItemTrace()

    def _setContent(self):

        definition = [
            '%sclass %s(%s, TypeDefinition):' %(ID1, self.getClassName(), self.sKlass),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%s%s' % (ID2, self.pnameConstructor()),
        ]
        if self.getPythonType() is None:
            definition.append('%s%s.__init__(self, pname, **kw)' %(ID3, self.sKlass))
        else:
            for l in self.getPyClassDefinition():
                definition.append(l)
            definition.append('%s%s.__init__(self, pname, pyclass=%s, **kw)' %(ID3, self.sKlass, self.getPyClass()))

        self.writeArray(definition)


class SimpleContentContainer(SimpleTypeContainer):

    type = DEF

    def setUp(self, tp):
        '''tp -- complexType/simpleContent/[Exention,Restriction]
        '''
        assert tp.isComplex() is True and tp.content.isSimple() is True,\
            'expecting complexType/simpleContent not: %s' %tp.content.getItemTrace()

        simple = tp.content
        dv = simple.content
        assert dv.isExtension() is True or dv.isRestriction() is True,\
            'expecting complexType/simpleContent/[Extension,Restriction] not: %s' \
            %tp.content.getItemTrace()

        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        self.content.attributeContent = dv.getAttributeContent()
        base = dv.getAttribute('base')
        if base is not None:
            self.sKlass = BTI.get_typeclass( base[1], base[0] )
            if not self.sKlass:
                self.sKlass   = base[1]
                self.sKlassNS = base[0]
            return

        raise Wsdl2PythonError,\
            'simple content derivation bad base attribute: ' %tp.getItemTrace()

    def _setContent(self):
        if type(self.sKlass) in (types.ClassType, type):
            definition = [
                '%sclass %s(%s, TypeDefinition):' \
                % (ID1, self.getClassName(), self.sKlass),
                '%s%s' % (ID2, self.schemaTag()),
                '%s%s' % (ID2, self.typeTag()),
                '%sattribute_typecode_dict = {}' % ID2,
                '%s%s' % (ID2, self.pnameConstructor()),
                ]
    
            if self.getPythonType() is None:
                definition.append('%s%s.__init__(self, pname, **kw)' %(ID3, self.sKlass))
            else:
                for l in self.getPyClassDefinition():
                    definition.append(l)
                definition.append(\
                    '%s%s.__init__(self, pname, pyclass=%s, **kw)'\
                    %(ID3, self.sKlass, self.getPyClass())
                )
        else:
            definition = [
                '%sclass %s(TypeDefinition):' % (ID1, self.getClassName()),
                '%s%s' % (ID2, self.schemaTag()),
                '%s%s' % (ID2, self.typeTag()),
                '%sattribute_typecode_dict = {}' % ID2,
                '%s%s' % (ID2, self.pnameConstructor()),
                '%s%s' % (ID3, self.nsuriLogic()),
                '%s'   % self.getBasesLogic(ID3),
                '%s%s.%s_Def.__init__(self, pname, **kw)' \
                % (ID3, NAD.getAlias(self.sKlassNS), self.sKlass),
                ]


        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))
        self.writeArray(definition)


class UnionContainer(SimpleTypeContainer):
    '''SimpleType Union
    '''
    type = DEF

    def __init__(self):
        SimpleTypeContainer.__init__(self)
        self.memberTypes = None

    def setUp(self, tp):
        if tp.content.isUnion() is False:
            raise ContainerError, 'content must be a Union: %s' %tp.getItemTrace()
        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        self.sKlass = 'ZSI.TC.Union'
        self.memberTypes = tp.content.getAttribute('memberTypes')

    def _setContent(self):
        definition = [
            '%sclass %s(%s, TypeDefinition):' \
            % (ID1, self.getClassName(), self.sKlass),
            '%smemberTypes = %s' % (ID2, self.memberTypes),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.pnameConstructor(self.sKlass)),
            ]
        self.writeArray(definition)


class ListContainer(SimpleTypeContainer):
    '''SimpleType List
    '''
    type = DEF

    def setUp(self, tp):
        if tp.content.isList() is False:
            raise ContainerError, 'content must be a List: %s' %tp.getItemTrace()
        self.name = tp.getAttribute('name')
        self.ns = tp.getTargetNamespace()
        self.sKlass = 'ZSI.TC.List'
        self.itemType = tp.content.getAttribute('itemType')

    def _setContent(self):
        definition = [
            '%sclass %s(%s, TypeDefinition):' \
            % (ID1, self.getClassName(), self.sKlass),
            '%sitemType = %s' % (ID2, self.itemType),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.pnameConstructor(self.sKlass)),
            ]
        self.writeArray(definition)


